from .tag import Tag, dataclass


@dataclass
class NTag(Tag):
    """ Namespaced tag that uses its class name as the prefix.

    Use it in a large application to have more control over prefixes
    """
    name: str
    __slots__ = 'name',

    PREFIX = ...

    def __init_subclass__(cls):
        cls.PREFIX = cls.__name__



import itertools
from typing import TypeVar, Mapping, Union, Iterable, List

from sqlalchemy.orm.base import instance_state
from sqlalchemy.orm.relationships import RelationshipProperty
from sqlalchemy.orm.state import InstanceState

from matroska_cache.dependency import PrimaryKey

SAInstanceT = TypeVar('SAInstanceT', bound=object)

# The dict used for plucking
PluckMap = Mapping[str, Union[int, 'PluckMap']]


def sa_dependencies(instance: Union[SAInstanceT, Iterable[SAInstanceT]], map: PluckMap, _seen: set = None) -> List[PrimaryKey]:
    """ Automatically collect PrimaryKey dependencies from SqlAlchemy instances

    Usage: when you have an output from a query, use sa_dependencies() on it to get a list of dependencies
    that you can provide to the MatroskaCache.put() function.
    NOTE: it will only pick primary key dependencies! Other dependencies can only be provided manually!

    Args:
        instance: the instance to get dependencies from
        map: inclusion map: {attribute: 1, relatiopnship: {key: 1, ...})
            Use `1` to include an attribute, dict() to include a relationship, `0` to exclude something
            NOTE: `map` must be valid!
    """
    if _seen is None:
        _seen = set()

    # Lists
    if isinstance(instance, (list, set, tuple)):
        return list(itertools.chain.from_iterable(
            sa_dependencies(item, map, _seen) for item in instance
        ))

    # Instances
    state: InstanceState = instance_state(instance)
    relationships: Mapping[str, RelationshipProperty] = state.mapper.relationships

    # Include self
    ret = []
    if instance not in _seen:
        ret.append(PrimaryKey(instance))
    _seen.add(instance)

    # If there's anything left to iterate, do it
    if isinstance(map, dict):
        for key, include in map.items():
            # Skip excluded elements
            if not include:
                continue

            # Skip non-relationships
            if key not in relationships:
                # TODO: implement dependencies on individual attributes?
                continue

            # Descend into the relationship
            ret.extend(sa_dependencies(getattr(instance, key), include, _seen))

    return ret



""" Matroska cache handles the caching of objects with dependency tracking.

Example:
    cache = MatroskaCache(RedisBackend(redis=self.redis, prefix='cache'))

    # Cache a piece of JSON data
    cache.put(
        # The cache key
        'users-list',
        # The data
        [
            {'id': 1, ...},
            {'id': 2, ...},
        ]
        # Dependencies. If any of them is updated, the whole cache key goes down
        dep.Id('user', 1),
        dep.Id('user', 2),
        # Cache time
        expires=60,
    )

    # Get the cache
    cache.get('users-list')

    # Invalidate the cache
    cache.invalidate(dep.Id('user', 1))

    # No cache anymore
    cache.has('users-list')  #-> False

When working with SqlAlchemy models, however, you may want to use some special tools :)
sa_dependencies() will automatically collect primary key dependencies:

Example:

    data = ssn.query(Article).options(
        joinedload(Article.author)
    ).all()

    dependencies = sa_dependencies(data, {'author': 1})
    # -> [
    #    PrimaryKey('Article', 1),
    #    PrimaryKey('Article', 2),
    #    PrimaryKey('Author', 1),
    # ]

    cache.put('articles-list', jsonify(data), *dependencies, expires=60)
"""

from typing import Any

from .backends.base import MatroskaCacheBackendBase
from .dependency.base import DependencyBase


class MatroskaCache:
    def __init__(self, backend: MatroskaCacheBackendBase):
        self.backend = backend

    def get(self, key: str) -> Any:
        """ Get cached data by `key`; raise KeyError if it does not exist

        Args:
            key: The cache key
        Raises:
            KeyError: no data cached by that key
        """
        return self.backend.get(key)

    def has(self, key: str) -> bool:
        """ Check if cache key `key` is available """
        return self.backend.has(key)

    def put(self, key: str, data: Any, *dependencies: DependencyBase, expires: int):
        """ Store data into the cache under key `key`

        Args:
            key: The cache key
            data: The data to store. It has to be json-serializable.
            *dependencies: List of dependencies for this cache entry. See `matroska_cache.dep`.
            expires: The number of seconds to keep this cache entry for
        """
        return self.backend.put(key, data, expires=expires, dependencies=dependencies)

    def delete(self, key: str):
        """ Delete a cache key """
        self.backend.delete(key)

    def invalidate(self, *dependencies: DependencyBase):
        """ Invalidate all cache entries that depend on `dependencies`

        Args:
            *dependencies: List of dependencies to invalidate cache records for
        """
        return self.backend.invalidate(dependencies)



from .base import DependencyBase, dataclass


@dataclass
class Tag(DependencyBase):
    """ Dependency on an arbitrary tag

    Usage:
        use this tag as a named signal to invalidate records

    Example:
        update_dashboard_for_admins = Tag('update-dashboard-for-admins')

        cache.put(
            'articles-list',
            [...],
            dependencies=[
                update_dashboard_for_admins,
                ...
            ]
        )

        cache.invalidate(update_dashboard_for_admins)
    """
    name: str
    __slots__ = 'name',

    PREFIX = 'tag'

    def key(self) -> str:
        return f'{self.PREFIX}:{self.name}'




__version__ = __import__('pkg_resources').get_distribution('matroska_cache').version


from .cache import MatroskaCache
from . import dependency as dep

try:
    from .sa_dependencies import sa_dependencies
except ImportError:
    pass



from typing import Union

from .base import DependencyBase, dataclass


@dataclass
class Id(DependencyBase):
    """ Dependency on an id of some object

    Usage:
        when the cached data includes an object, use its id to declare a dependency

    Example:
        cache.put(
            'articles-list',
            [
                Article(id=1).asdict(),
                Article(id=2).asdict(),
            ],
            dependencies=[
                Id('article', 1),
                Id('article', 2),
            ]
        )

        cache.invalidate(
            Id('article', 1)
        )
    """
    type: str
    id: Union[int, str]
    __slots__ = 'type', 'id'

    PREFIX = 'id'

    def key(self) -> str:
        return f'{self.PREFIX}:{self.type}:{self.id}'



import itertools
import json
from typing import Any, Iterable, List

from redis import Redis

from matroska_cache.dependency.base import DependencyBase
from .base import MatroskaCacheBackendBase


class RedisBackend(MatroskaCacheBackendBase):
    def __init__(self, redis: Redis, *, prefix: str):
        """ Init the Redis backend for the matroska cache

        Args:
            redis: Redis client
            prefix: Prefix string for our cache keys
        """
        self.redis = redis
        self.prefix = prefix

    def get(self, key: str) -> Any:
        # Get the data; fail if the key does not exist
        data = self.redis.get(self._key('data', key))
        if data is None:
            raise KeyError(key)

        # Unserialize
        return unserialize(data)

    def has(self, key: str) -> bool:
        return self.redis.exists(self._key('data', key)) == 1

    def delete(self, key: str):
        self.redis.delete(self._key('data', key))

    def put(self, key: str, data: Any, dependencies: Iterable[DependencyBase], expires: int):
        # Store the dependency information
        self._remember_dependencies_for(key, dependencies, expires)

        # Store the data
        self.redis.setex(self._key('data', key), expires, serialize(data))

    def invalidate(self, dependencies: Iterable[DependencyBase]):
        if not dependencies:
            return

        # Get the list of keys that depend on `dependency` (every single one of them)
        # This is resolved through the `rdep` key
        rdep_keys = [self._key('rdep', dependency.key())
                     for dependency in dependencies]
        p = self.redis.pipeline()
        for rdep_key in rdep_keys:
            p.smembers(rdep_key)
        data_keys = list(itertools.chain(*p.execute()))

        # Invalidate every data key that depend on these rdep_keys
        # This means deleting them
        if data_keys:  # do nothing if there is nothing to do
            self.redis.delete(*(self._key('data', key) for key in data_keys))

    def _remember_dependencies_for(self, key: str, dependencies: Iterable[DependencyBase], expires: int):
        """ Update dependency information for `key`

        This method stores two keys:
        * fdep: forward dependency information: { data-key => set(dependency, ...) }
        * rdep: reverse dependency information: { dependency => set(data-key, ...) }

        After storing those, it updates the expiration time on every dependency key:
        sets it to `expires`, but makes sure that the resulting TTL is not getting shorter
        """
        # Dependencies as a string
        deps = [dependency.key() for dependency in dependencies]
        if not deps:
            return

        # Use a pipeline to speed up
        p = self.redis.pipeline()

        # Forward dependency information: `data` depends on `dep`
        fdep_key = self._key('fdep', key)
        p.sadd(fdep_key, *deps)

        # Reverse dependency information: `dep` is a dependency of `data`
        rdep_keys = []
        for dep in deps:
            rdep_key = self._key('rdep', dep)
            rdep_keys.append(rdep_key)
            p.sadd(rdep_key, key)

        # Commit
        p.execute()

        # Get the current TTLs for dependency keys
        dep_keys = [fdep_key, *rdep_keys]
        for key in dep_keys:
            p.ttl(key)
        dep_keys_ttls: List[int] = p.execute()

        # Prolong every TTL if `expires` is beyond that.
        dep_keys_expire = [max(ttl, expires) for ttl in dep_keys_ttls]

        # Update TTLs for dep keys
        for key, key_expires in zip(key, dep_keys_expire):
            p.expire(key, key_expires)
        p.execute()

    def _key(self, type: str, name: str):
        """ Make a Redis key name

        Args:
            type: The type of information stored in the key.
                'data': the data cached by the user
                'fdep': forward dependencies
            name: Cache key
        """
        return f'{self.prefix}::{type}::{name}'


serialize = json.dumps
unserialize = json.loads



from abc import ABC, abstractmethod
from dataclasses import dataclass


@dataclass
class DependencyBase(ABC):
    # Performance
    __slots__ = ()

    # key() prefix. Used for overrides in subclasses
    PREFIX = '?'

    @abstractmethod
    def key(self) -> str:
        """ Get string representation of this dependency """



from typing import Tuple, Any

from sqlalchemy.orm.base import instance_state
from sqlalchemy.orm.state import InstanceState

from .id import Id, dataclass


@dataclass()
class PrimaryKey(Id):
    """ Dependency on a primary key of an instance

    Usage:
        when the cached data includes an SqlAlchemy instance, use its primary key to declare a dependency

    Example:
        cache.put(
            'articles-list',
            [
                Article(id=1).asdict(),
                Article(id=2).asdict(),
            ],
            dependencies=[
                Id('article', 1),
                Id('article', 2),
            ]
        )

        cache.invalidate(
            Id('article', 1)
        )
    """

    PREFIX = 'pk'

    def __init__(self, instance: object):
        state: InstanceState = instance_state(instance)
        super().__init__(
            type=state.class_.__name__,
            id=self._instance_identity_to_str(state.identity)
        )

    def _instance_identity_to_str(self, identity: Tuple[Any]) -> str:
        if len(identity) == 1:
            return str(identity[0])
        else:
            return repr(identity)


@dataclass()
class RawPrimaryKey(Id):
    """ A primary key dependency that you will have to construct manually ;) """
    PREFIX = 'pk'



""" Cache back-end for Matroska Cache.

It's primary task is to handle dependencies when the following operations are made:

* put(data, key, dependencies)
* get(key)
* invalidate(dependencies)

Example put():
    put(
        data=[
            Article(id=1, author=User(12)).asdict(),
            Article(id=2, author=User(12)).asdict(),
        ],
        key='articles-list',
        dependencies=[
            Id('article', 1),
            Id('article', 2),
            Id('author', 12),
        ],
    )

Example get():
    get('articles-list')

Example invalidate:
    invalidate(
        dependencies=[
            Id('author', 12),
        ]
    )
"""

from abc import ABC, abstractmethod
from typing import Any, Iterable

from ..dependency.base import DependencyBase


class MatroskaCacheBackendBase(ABC):
    @abstractmethod
    def put(self, data: Any, key: str, *, expires: int, dependencies: Iterable[DependencyBase]):
        """ Put `data` into cache, keyed by `key`, depending on `dependencies`

        Args:
            data: The data to put into cache
            key: The key under which caching is done. Has to be unique.
            dependencies: The list of objects that the data in cache depends on.
                If any of those dependencies becomes invalid, this piece of data will be invalidated as well.
            expires: The number of seconds the data will expire in
        """

    @abstractmethod
    def get(self, key: str) -> Any:
        """ Get a piece of cached data, if still available

        Args:
            key: The cache key
        Returns:
            cached data
        Raises:
            KeyError: no data found
        """

    @abstractmethod
    def has(self, key: str) -> bool:
        """ See whether the cache has `key` stored in it """

    @abstractmethod
    def delete(self, key: str):
        """ Remove cached data by key """

    @abstractmethod
    def invalidate(self, dependencies: Iterable[DependencyBase]):
        """ Invalidate all cache records with `dependency` as their dependency """



from __future__ import annotations

import warnings
from typing import Any, List, Callable, MutableMapping, Tuple

from .base import DependencyBase, dataclass
from .tag import Tag

ExtractorFunc = Callable[[Any], dict]


class Scopes:
    """ Generate dependencies that describe lists of objects.

    This tool is designed to solve the case where newly created, or freshly removed, items may enter the scope
    of some listing, which may itself be filtered by some condition.

    In general, if you cache a list of objects by `Id`:

        cache.put(
            'articles-list', [...],
            dep.Id('article', 1),
            dep.Id('article', 2),
            ...
        )

    you will not have this list of articles invalidated when new articles come into scope.

    For instance, if your view caches the list of articles by `category`, intialize the `Scopes` object like this:

        article_scopes = Scopes('article', production_mode=False)

        @article_scopes.describes('category')
        def article_category(article: Article):
            return {'category': article.category}

    This operation enables us to use `category` information as a dependency for the cached list:

        # Articles filtered by category ...
        articles = ssn.query(Article).filter_by(category='python').all()
        cache.put(
            f'articles-list:category=python',  # make sure to put it here
            articles,
            ...
            # ... declare this category as their dependency
            article_scopes.condition(category='python')
        )

    Now, in another place, where articles are created, you can invalidate this dependency automatically
    by just passing the new article to the article_scopes.invalidate_for() method:

        def create_new_article(...):
            ...
            article_scopes.invalidate_for(article, invalidate_for)

    Under the hood, it will go over every condition known through @article_scopes.describes()
    and invalidate all related caches.

    ---

    NOTE: Does it seem complicated to you?
    It is; but this complexity follows one goal: to make caching *declarative* and minimize hidden connections in your code.
    For instance, you could have used Tag() to achieve the very same result.
    Listing articles:

        articles = ssn.query(Article).filter_by(category='python').all()
        cache.put(
            f'articles-list:category=python',  # make sure to put it here
            articles,
            ...
            # ... declare this category as their dependency
            dep.Tag(f'articles:category=python'),
        )

    Adding articles:

        cache.invalidate(dep.Tag(f'books:category={article.category}'))

    This code would work just fine; but then, for every caching behavior you would need *to remember* to add another line
    to the place where articles are saved. Those connections would soon become numerous and lead to caching errors that
    are hard to catch.

    This approach with `Scopes()` is a declarative approach:
    you first declare *the intention* of caching by category, and `Scopes()` will check that everything is set up properly.
    """
    def __init__(self, object_type: str, *, production_mode: bool):
        """ Initialize scopes for a particular kind of object

        Args:
            object_type: Name for the objects you're watching. Got to be unique. Example: 'article'
            production_mode: Whether the cache is currently operating on a production server.
                If there is an error with how you configured the `Scopes` object, its will be disabled.
                In development (production_mode=False), an exception will be raised.
        """
        self._object_type = object_type
        self._extractor_fns: MutableMapping[Tuple[str], ExtractorFunc] = {}

        # An invalidate-all dependency used to invalidate all caches in cases when scopes are not used properly.
        # For instance, the user is attempting to cache the results that matched a filter
        #   .condition(category_id=10)
        # but there was no extractor function that describes how `category_id` influences the cache.
        self._invalidate_all = InvalidateAll(self._object_type)
        self._production_mode = production_mode

    def describes(self, *param_names):
        """ Decorator for a function that extracts data for a conditional dependency.

        Whenever a new object is created, your application should call `invalidate_for()`,
        and it will invalidate every cache that depends on any filtered list of such objects.

        The arguments for the scope are described by the decorated function: if you want to cache the results of
        a list filtered by `category=<something>`, you first need to define an extractor function:

            @article_scopes.describes('category')
            def article_category(article: Article, **info):
                # Extract filter arguments from a new object
                return {'category': article.category}

        Only after such a condition is described, you can use it as a cache key:

            cache.put(
                f'articles-{category}',
                articles,
                ...,
                *article_scopes.condition(category=category),
                expires=600,
            )

        Note that the values extracted by `article_category()` and provided to `condition()` have to match.
        If they don't, cache will misbehave.

        Args:
            *param_names: The list of parameter names the extractor function is going to return.
                These names are completely custom, but have to match those given to condition()
        """
        def decorator(fn: ExtractorFunc):
            """ Register the decorated function and return """
            # Remember the extractor function and the parameters it returns
            filter_params_signature = tuple(sorted(param_names))
            assert filter_params_signature not in self._extractor_fns, (
                f'An extractor with filter parameters {filter_params_signature!r} is already described'
            )
            self._extractor_fns[filter_params_signature] = fn

            # Done
            return fn
        return decorator

    def invalidate_for(self, item: Any, cache: 'MatroskaCache', **info):
        """ Invalidate all caches that may see `item` in their listings.

        Args:
            item: The new/deleted item that may enter or leave the scope of some listing
            cache: MatroskaCache to invalidate
            **info: Extra info that may be passed to your extractor functions
        """
        cache.invalidate(*self.object_invalidates(item, **info))

    def condition(self, **conditions: Any) -> List[ConditionalDependency]:
        """ Get dependencies for a conditional scope.

        Use this method with MatroskaCache.put() to generate dependencies for your scope.

        Args:
            **conditions: The description of your filtering conditions, in the `name=value` form.

        Returns:
            List of scope dependencies to be used on your cache entry
        """
        # Signature
        filter_params_signature = tuple(sorted(conditions))

        if filter_params_signature in self._extractor_fns:
            return [
                ConditionalDependency(self._object_type, conditions),
                # Got to declare this kill switch as a dependency; otherwise, it won't work.
                self._invalidate_all,
            ]
        elif self._production_mode:
            warnings.want(
                f'Matroska cache: no extractor @describes for {filter_params_signature!r}. '
                f'Caching disabled. '
            )
            return [self._invalidate_all]
        else:
            raise RuntimeError(
                f'No extractor function is described for condition {filter_params_signature!r}. '
                f'Please use @.describes() on a function with matching parameters. '
                f'It will not fail in production, but caching will be disabled.'
            )

    def object_invalidates(self, item: Any, **info) -> List[ConditionalDependency]:
        """ Get dependencies that will invalidate all caches that may see `item` in their listings.

        This function takes the `item` and calls every extractor function decorated by `@scope.describes()`.
        The resulting value will be used to find scopes that this object will come into, and invalidate them.

        Args:
            item: The newly created or freshly deleted item.
            **info: Additional arguments to pass to *all* the extractor functions.

        Returns:
            List of dependencies to be used with `cache.invalidate()`
        """
        ret = []
        for param_names, extractor_fn in self._extractor_fns.items():
            params = extractor_fn(item, **info)
            if set(params) == set(param_names):
                ret.append(ConditionalDependency(self._object_type, params))
            elif self._production_mode:
                ret.append(self._invalidate_all)
            else:
                raise RuntimeError(
                    f'The described extractor {extractor_fn} was supposed to return a dict of {param_names!r}, '
                    f'but it returned only {params!r}. Please fix. '
                    f'It will not fail in production, but caching will be disabled.'
                )
        return ret


@dataclass
class ConditionalDependency(DependencyBase):
    """ Internal dependency used by Scope

    A dependency object of this type is generated for the output of every extractor function.
    This is how the whole thing operates:

    When a new article is created, it is passed to the `invalidate_for()` function.
    An extractor function, described like this:

            @article_scopes.describes('category')
            def article_category(article: Article, **info):
                # Extract filter arguments from a new object
                return {'category': article.category}

    will generate a dependency:

        ConditionalDependency(object_type='article', conditions={'category': 'sci-fi'})
        # it is just a string:
        'condition:article:&category=sci-fi&'

    This string invalidates any cache entries that had been created like this:

            cache.put(
                ...
                *article_scopes.condition(category=category),
            )

    So, in essense, this whole Scopes is just an interface to match the two strings in a declarative fashion.
    """
    object_type: str
    condition: str
    __slots__ = 'object_type', 'condition',

    def __init__(self, object_type: str, conditions: dict):
        self.object_type = object_type
        self.condition = '&'.join(f'{key}={value}'
                                  # items are sorted to make sure they always match in the same way!
                                  for key, value in sorted(conditions.items()))
        # Surround it with &s to enable wildcard matching
        self.condition = '&' + self.condition + '&'

    PREFIX = 'condition'

    def key(self) -> str:
        return f'{self.PREFIX}:{self.object_type}:{self.condition}'


class InvalidateAll(Tag):
    """ A custom tag, used in production, to invalidate all scopes in cases when Scopes is misconfigured """

    # Use the same prefix. Not important; just looks nice
    # There will be no clashes because all `ConditionalDependency` have "&" in their names
    PREFIX = ConditionalDependency.PREFIX

    def __init__(self, object_type: str):
        super().__init__(f'{object_type}')
